#ifndef ROOTFINDER_H
#define ROOTFINDER_H

#include "rootcnt.hpp"
#include <list>

template<typename F, typename F1, typename F2, typename Real = double>
class RootFinder {
   public:
      RootFinder(F f_, F1 f1_, F2 f2_) : f(f_), f1(f1_), f2(f2_) {
      }

      template<typename OutputIterator>
      void get_roots(OutputIterator outit,
	    Real a, Real b, Real eps) const {
	 unsigned int numOfRoots = get_count(a, b);
	 if (numOfRoots > 0) {
	    get_roots(outit, a, b, eps, numOfRoots);
	 }
      }

   private:
      F f; F1 f1; F2 f2;

      struct Task {
	 Task(Real a_, Real b_, unsigned int numOfRoots_) :
	    a(a_), b(b_), numOfRoots(numOfRoots_) {
	 }
	 Task() : a(0), b(0), numOfRoots(0) {
	 }
	 Real a, b;
	 unsigned int numOfRoots;
      };

      unsigned int get_count(Real a, Real b) const {
	 return get_root_count(a, b, f, f1, f2);
      }

      static int sgn(Real x) {
	 return x < 0? -1: (x > 0? 1: 0);
      }

      Real bisection(Real a, Real b, Real eps) const {
	 Real fa = f(a); Real fb = f(b);
	 for(;;) {
	    Real midpoint = (a + b) / 2; Real fm = f(midpoint);
	    if (fm == 0 || b - a < eps) return midpoint;
	    if (sgn(fa) * sgn(fm) < 0) {
	       b = midpoint; fb = fm;
	    } else {
	       a = midpoint; fa = fm;
	    }
	 }
      }

      template<typename OutputIterator>
      void get_roots(OutputIterator& outit,
	       Real a, Real b, Real eps, unsigned int numOfRoots) const {
	 std::list<Task> tasks; // shared
	 unsigned int numOfRootsFound = 0; // shared

	 if (numOfRoots == 0) return;
	 tasks.push_back(Task(a, b, numOfRoots));
#pragma omp parallel
	 while (numOfRootsFound < numOfRoots) {
	    // fetch next task, if there is any
	    Task task;
	    bool busyloop = false;
#pragma omp critical
	    if (tasks.size() > 0) {
	       task = tasks.front(); tasks.pop_front();
	    } else {
	       busyloop = true;
	    }
	    if (busyloop) {
#pragma omp taskyield
	       continue;
	    }

	    // now we are working on a task
	    if (task.numOfRoots == 0) continue;
	    if (task.numOfRoots == 1) {
	       Real root = bisection(task.a, task.b, eps);
#pragma omp critical
	       {
		  *outit++ = root; ++numOfRootsFound;
	       }
	       continue;
	    }
	    Real midpoint = (task.a + task.b) / 2;
	    unsigned int numOfLeftRoots = get_count(task.a, midpoint); 
	    unsigned int numOfRightRoots = get_count(midpoint, task.b);
	    if (numOfLeftRoots + numOfRightRoots < task.numOfRoots) {
#pragma omp critical
	       {
		  *outit++ = midpoint; ++numOfRootsFound;
	       }
	    }
#pragma omp critical
	    {
	       tasks.push_back(Task(task.a, midpoint, numOfLeftRoots));
	       tasks.push_back(Task(midpoint, task.b, numOfRightRoots));
	    }
	 }
      }
};

template<typename OutputIterator, typename F, typename Real = double>
void get_roots(OutputIterator outit, F f, Real a, Real b, Real eps) {
   NumF1<F, Real> f1(f);
   NumF2<F, Real> f2(f);
   RootFinder<F, decltype(f1), decltype(f2), Real> rfinder(f, f1, f2);
   rfinder.get_roots(outit, a, b, eps);
}

template<typename OutputIterator,
   typename F, typename F1, typename F2, typename Real = double>
void get_roots(OutputIterator outit, F f, F1 f1, F2 f2,
      Real a, Real b, Real eps) {
   RootFinder<F, F1, F2, Real> rfinder(f, f1, f2);
   rfinder.get_roots(outit, a, b, eps);
}

#endif
